//
// Avi2Cvc D by ssjx (http://ssjx.co.uk)
//
// Compile with:	
//		dmd -i avi2cvc.d
//
// Version 2.8 D (25/02/23)
// 	25/02/2023 - Source clean up, now Avi2Cvc/Bmp2Cvi share a lot of functions
//
// Version 2.7 D (16/02/23)
//
//
//
import std.stdio;
import std.file;
import std.bitmanip;	// swapEndian
import std.math;		// abs

import cv1,msvc,utils;

// 56 
struct Avih{	
	align(1):
	int microps;
	int maxbps;
	int res;
	int flags;
	int totf;
	int initf;
	int streams;
	int buffersize;
	int biWidth;
	int biHeight;
	int scale;
	int fps;
	int strt;
	int leng;
};

//
struct Vsdhr{	
	align(1):
	char[4] subtype;
	char[4] handler; // e.g. MSVC
	int flags;
	int pri;
	int init;
	int scale;
	int fps;
	int strt;
	int lng;
	int bsize;
	int quality;
	int ssize;
	short[4] rectframe;
};

// 40
struct Vstrf{	
	align(1):
	int biSize;
	int biWidth;
	int biHeight;
	short biPlanes;
	short biBitCount;
	int biCompression;
	int biSizeImage;
	int biXPelsPerMeter;
	int biYPelsPerMeter;
	int biClrUsed;
	int biClrImportant;
};

// 8 
struct Hdr{	
	align(1):
	char[4] db;
	int size;
};

// ---
Avih[1] avi_header;
Vsdhr[1] str_header;
Vstrf[1] str_format;
Hdr[1] header;	
// ---

My_cvc_header[1] cvc_header;
My_cvc_video[1] cvc_video;
My_cvc_frame[1] cvc_frame;

//
struct Newchunk{	
	align(1):
	int size;
	char[4] chunktype;
};

Newchunk[1] nchunk;

//
//
//

ubyte[16000] gray;
ubyte[4000] pic;

void main(string[] args)
{
	// writeln("Avi Header: ",avi_header.sizeof);
	// writeln("Str Header: ",str_header.sizeof);
	// writeln("Str Format: ",str_format.sizeof);

	int dither=1;
	string fn;
	
	writeln("Avi2Cvc v2.8 D (25/02/23) by ssjx (http://ssjx.co.uk)");
	writeln("-----------------------------------------------------");
	
	if (args.length>1){
		foreach(par;args){
			if (par[0]=='-'){
				if (par=="-nodither"){dither=0;}
				if (par=="-ordered"){dither=1;}
				if (par=="-diffuse"){dither=2;}
			}else{
				if (par!=args[0]){
					// Not the exe or an option so probably an avi file
					fn=par;
				}
			}
		}
	}else{
		writeln("Usage: avi2cvc [option] myfile.avi");
		writeln("Options: -nodither , -ordered (default), -diffuse");
		writeln("Avi should be 160x100 and either uncompressed or MSVC compressed.");
		return;
	}

	// Start reading AVI file
	if (exists(fn)){	
		immutable string outfn=fn~".cvc";	// <-- Note squiggle used to join strings!
	
		writeln("In: ",fn);
		writeln("Out: ",outfn);
		writeln("Dither: ",dither);
	
		try{
			bool ismovie=false;	
			File file= File(fn, "rb");
	    
			file.seek(12);	
			//
			// New AVI parser
			// 
			char[4] avi_list;
			char[4] subtype;
			char[4] handler;
			
			while(true) {
				int size;
				file.rawRead(avi_list);
				
				char[4] ctype=avi_list;

				if (ctype=="LIST") {
					file.rawRead(nchunk);
					size = nchunk[0].size;
					// ctype = to!string(nchunk[0].chunktype);
					
					ctype=nchunk[0].chunktype;
					writeln("List : ",ctype);

					switch (ctype){
					case "movi":
						 writeln("Movie List!");
						ismovie = true;
					break;
					case "hdrl":
						 writeln("Header List");
					break;
					case "strl":
						 writeln("Stream List");
					break;
					default:
						 writeln("Unknown List ",ctype);
					break;
					}
				} else {
					// Chunk
					int[1] avi_chunksize;
					file.rawRead(avi_chunksize);
					size = avi_chunksize[0];	

					switch (ctype){
					case "avih":
						// AVI file header
						writeln("+ Avi Header ",size);
						file.rawRead(avi_header);
						int bmhd_frames=avi_header[0].totf;
						writeln("  Frames: ", bmhd_frames);
					break;
					
					case "strh":
						// Stream header
						writeln("+ Stream Header ",size);
						file.rawRead(str_header);
						// subtype = to!string(str_header[0].subtype);
						subtype = str_header[0].subtype;

						if (subtype=="vids"){
							
							handler=str_header[0].handler;
							
							// Upper to lower case
							foreach(ref h;handler){
								if (h>=65 && h<=90){h+=32;}
							}
							
							writeln("  Format: [",handler,"]");

							int scale=str_header[0].scale;
							int fps=str_header[0].fps;
							int bmhd_fps = fps / scale;

							writeln("  Rate: ", fps, " / ", scale, " [", bmhd_fps, "]");

						} else {
							 writeln("  Ignoring: ",subtype);
						}
					break;
					
					case "strf":
						// Stream format
						writeln("+ Stream Format ",size);
						file.rawRead(str_format);

						if (subtype=="vids"){
							int bmhd_w = str_format[0].biWidth;
							int bmhd_h = str_format[0].biHeight;
							int bmhd_d = str_format[0].biBitCount;
							int bmhd_bytes = bmhd_d / 8;

							 writeln("  Size: ", bmhd_w, " x ", bmhd_h, " x ", bmhd_d);
						} else {
							 writeln("  Ignoring: ",subtype);
						}
					break;
					
					case "JUNK":
						writeln("+ Junk ",size);
						file.seek(size,SEEK_CUR);	
					break;
					
					default:
						writeln("+ Unknown ", ctype, " Size:", size);
						file.seek(size,SEEK_CUR);
					break;
					}
				}

				if (size==0){
					 writeln("Panic Exit!");
					ismovie = false;
					break;
				}

				if (ismovie==true){
					break;
				}
			}
							
			//
			if ((handler!= "dib ") && (handler!="msvc")) {
				writeln("Unsupported compression [", handler, "], video must be uncompressed DIB or MSVC.");
				return;
			}
		
			// Make sure AVI is correct resolution
			if ((str_format[0].biWidth > 160) || ( str_format[0].biHeight > 100)) {
				writeln("Resolution must be 160x100 or lower.");
				return;
			}

			if ((str_format[0].biWidth < 0) || ( str_format[0].biHeight < 0)) {
				writeln("Resolution must be 160x100 or lower.");
				return;
			}

			if ((str_format[0].biWidth % 4) > 0) {
				writeln("AVI width must be divisible by 4");
				return;
			}
			
			//

			if (ismovie==true){
				cvc_header[0].id[0]='c';
				cvc_header[0].id[1]='y';
				cvc_header[0].id[2]='v';
				cvc_header[0].id[3]='d';	
				
				cvc_header[0].size=0;
				cvc_header[0].lock=0;
				cvc_header[0].headers=cast(short)1.swapEndian;
				
				cvc_video[0].id[0]='v';
				cvc_video[0].id[1]='i';
				cvc_video[0].id[2]='d';
				cvc_video[0].id[3]='o';
				
				short w=cast(short)str_format[0].biWidth;
				short h=cast(short)str_format[0].biHeight;
				
				cvc_video[0].w=w.swapEndian;
				cvc_video[0].h=h.swapEndian;
				cvc_video[0].frames=avi_header[0].totf.swapEndian;
				
				short fps=cast(short)(str_header[0].fps/str_header[0].scale);
				cvc_video[0].fps=fps.swapEndian;
				
				cvc_video[0].format[0]='c';
				cvc_video[0].format[1]='y';
				cvc_video[0].format[2]='v';
				cvc_video[0].format[3]='1';
				
				try
				{
					ubyte[] cvc_data;
					ubyte[] bmp;
					ubyte[] frame;
					
					ubyte[16000] gray8;	// 8bit gray frame
					
					// Create empty variable sized previous frame
					ubyte[] prev_frame;
					prev_frame.length=pic.length;
					prev_frame[]=0;
					set_as_previous(prev_frame);
					//
					
					File outfile = File(outfn, "wb");   	
			        writeln("Saving as ",outfn,"...");
					
					outfile.rawWrite(cvc_header);
					outfile.rawWrite(cvc_video);
					
					//
					// short maxsize=cast(short)((str_format[0].biWidth*str_format[0].biHeight)/4);
					int maxsize=(str_format[0].biWidth*str_format[0].biHeight)/4;
					
					foreach(i;0..avi_header[0].totf){
						// Read a frame from the AVI
						file.rawRead(header);
						
						final switch(handler)
						{
						case "msvc":
							frame.length=header[0].size;
							file.rawRead(frame);
							decode_msvc(frame,gray8,str_format[0].biWidth,str_format[0].biHeight);
							
							gray=gray8;
						break;
						
						case "dib ":
							bmp.length=header[0].size;	// May include padding so best not use x*y*depth
							file.rawRead(bmp);
							grey(bmp,gray,str_format[0].biWidth,str_format[0].biHeight);
						break;
						}
						
						switch(dither)
						{
						case 0:
							twobit(gray,str_format[0].biWidth,str_format[0].biHeight);
						break;
						
						case 1:
							ordered(gray,str_format[0].biWidth,str_format[0].biHeight);
						break;
						
						case 2:
							diffuse(gray,str_format[0].biWidth,str_format[0].biHeight);
						break;
						
						default:
							ordered(gray,str_format[0].biWidth,str_format[0].biHeight);
						break;
						}
						
						// 8bit > 2bit
						frameconv(gray,pic,str_format[0].biWidth,str_format[0].biHeight);
					
						// See which compression method works the best!
						short esize=encode_rle_size(pic);
						short usecomp=1;	
						short tmpsize=0;
					
						tmpsize=encode_delta1_size(pic);
						if (tmpsize<esize){
							esize=tmpsize;
							usecomp=2;
						}
						
						tmpsize=encode_delta2_size(pic);
						if (tmpsize<esize){
							esize=tmpsize;
							usecomp=3;
						}
					
						// In case compression just does not work (or not worth it)...
						if (esize>=maxsize){
							esize=cast(short)maxsize;
							usecomp=0;
						}
						
						// Force
						// esize=cast(short)maxsize;					
						// usecomp=0;
						
						cvc_data.length=esize;
						
						final switch(usecomp)
						{
						case 0:
							// Uncompressed frame
							encode_none(pic,cvc_data);
						break;
						
						case 1:
							// RLE compressed frame...
							encode_rle(pic,cvc_data);
						break;
						
						case 2:
							// Delta 1 compressed frame...
							encode_delta1(pic,cvc_data);
						break;
						
						case 3:
							// Delta 2 compressed frame...
							encode_delta2(pic,cvc_data);
						break;
						}
					
						// Write  the CVC frame
						cvc_frame[0].type=0;
						cvc_frame[0].subtype=usecomp.swapEndian;
						cvc_frame[0].size=esize.swapEndian;
						
						writeln("Frame ",i,": ",usecomp," - ",pic.length," => ",esize);
						
						outfile.rawWrite(cvc_frame);
						outfile.rawWrite(cvc_data);	
					}
					
				}catch (FileException e){
				    writeln("Exception caught creating output file!");
				}
			}

		}catch (FileException e){
		    writeln("Exception caught opening input file!");
		}
	}else{
		writeln("Could not find the file ",fn);
	}
}