//
// Encoding functions shared by both Avi2Cvc (Video) and Bmp2Cvi (Image) converters
//

module cv1;
// import std.stdio;

@safe:

// Main header <12>
struct My_cvc_header {
	align(1):
	char[4] id; 		// cyvd
	int size;
	short lock;			// encryption key 
	short headers;		// number of headers to expect
};

// Video <18> 
struct My_cvc_video{
	align(1):
	char[4] id; 		// "vido"
	short w;	
	short h;
	int frames;
	short fps;
	char[4] format;    	// "cyv1"
};

// Frame header <6>
struct My_cvc_frame{
	align(1):
	short type;
	short size;
	short subtype;
};

private ubyte []prev;

void set_as_previous(ubyte[] pix){
	// Was a loop now it's just this...
	prev.length=pix.length;
	prev[]=pix[];
}

// 0:Uncompressed
int encode_none(ubyte[] pix, ubyte[] outdata){
	foreach(i;0..pix.length){
		outdata[i]=pix[i];
	}

	set_as_previous(pix);
	return pix.length;
}

// 1: Runlength encoding
int encode_rle(ubyte[] pix, ubyte[] outdata){
	int pos;
	ubyte pb=pix[0];
	ubyte cnt=1;

	foreach(i;1..pix.length){
		if ((pix[i]==pb) && (cnt<126)){
			cnt++;
		}else{
			outdata[pos]=cnt;
			outdata[pos+1]=pb;
			pos+=2;
		
			pb=pix[i];
			cnt=1;
		}
	}
	
	if (cnt!=0){
		outdata[pos]=cnt;
		outdata[pos+1]=pb;
		pos+=2;
	}
	
	set_as_previous(pix);
	return pos;
}


short encode_rle_size(ubyte[] pix){
	short pos;
	ubyte pb=pix[0];
	ubyte cnt=1;

	foreach(i;1..pix.length){
		if ((pix[i]==pb) && (cnt<126)){
			cnt++;
		}else{
			pos+=2;
			pb=pix[i];
			cnt=1;
		}
	}
	
	if (cnt!=0){pos+=2;}
//	writeln("RLE Size:",pos);
	return pos;
}

// 2: Delta1 - RLE + Skipping same as previous frame pixels
int encode_delta1(ubyte[] pix,ubyte[] outdata){
	int cb,pb;
	int pos;
	int no=1;

	pb=pix[0];
	if (pix[0]==prev[0]){pb=-1;}

	foreach(i;1..pix.length){
		cb=pix[i];
		if (pix[i]==prev[i]){cb=-1;}

		if (cb==pb && no<126){
			no++;
		}else{
			if (pb==-1){
				outdata[pos++]=cast(ubyte)-no;
			}else{
				outdata[pos++]=cast(ubyte)no;
				outdata[pos++]=pix[i-1];
			}

			no=1;
		}

		pb=cb;
	}

	if (no!=0){
		if (pb==-1){
			outdata[pos++]=cast(ubyte)(-no);
		}else{
			outdata[pos++]=cast(ubyte)no;
			outdata[pos++]=pix[pix.length-1];	// Last pixel
		}
	}

	set_as_previous(pix);
	return pos;
}


short encode_delta1_size(ubyte[] pix){
	short pos;
	int cb,pb,no=1;

	pb=pix[0];
	if (pix[0]==prev[0]){pb=-1;}

	foreach(i;1..pix.length){
		cb=pix[i];
		
		if (pix[i]==prev[i]){cb=-1;}

		if (cb==pb && no<126){
			no++;
		}else{
			if (pb==-1){
				pos++;
			}else{
				pos+=2;
			}
			no=1;
		}

		pb=cb;
	}

	if (no!=0){
		if (pb==-1){
			pos++;
		}else{
			pos+=2;
		}
	}
//writeln("Delta 1 Size:",pos);
return pos;
}

//3: Delta 2 - Byte copy skipping pixel same as previous frame
int encode_delta2(ubyte[] pix,ubyte[] outdata){
	int pos,aw,no,xo;

	foreach(i;0..pix.length){
		if (pix[i]==prev[i]){
			if (xo>0){
				outdata[pos++]=cast(ubyte)xo;
				foreach(c;0..xo){
					outdata[pos++]=pix[i-xo+c];
				}
				xo=0;
			}

			no++;
			if (no>126){
				outdata[pos++]=cast(ubyte)-no;
				no=0;
			}
		}else{
			xo++;
			if (no>0){
				outdata[pos++]=cast(ubyte)-no;
				no=0;
			}

			if (xo>=126){
				outdata[pos++]=cast(ubyte)xo;
				foreach(c;0..xo){
					outdata[pos++]=pix[i-xo+c+1];
				}
				xo=0;
			}
		}
	}

	//
	if (no>0){
		outdata[pos++]=cast(ubyte)-no;
	}

	if (xo>0){
		outdata[pos++]=cast(ubyte)xo;
		foreach(c;0..xo){
			// Should this be a 'datasize-1' ???? Bmp2Cvi test says no..
			outdata[pos++]=pix[pix.length-xo+c];
		}
	}

	set_as_previous(pix);
	return pos;
}


short encode_delta2_size(ubyte[] pix){
	short pos;
	int no,xo;

	foreach(i;0..pix.length){
		if (pix[i]==prev[i]){
			if (xo>0){
				pos+=(1+xo);
				xo=0;
			}

			no++;

			if (no>126){
				pos++;
				no=0;
			}
		}else{
			xo++;

			if (no>0){
				pos++;
				no=0;
			}

			if (xo>=126){
				pos+=(1+xo);
				xo=0;
			}
		}
	}

	//
	if (no>0){pos++;}
	if (xo>0){pos+=(1+xo);}

	//writeln("Delta 2 Size:",pos);

	return pos;
}
